"""Save all documents."""

from __future__ import annotations

try:
    import docxrev
except TypeError as error:
    raise TypeError(
        "Cannot access the document. Save the document, close it,\n"
        "then re-open it and try again."
    ) from error

from gradedoc import shared


def save_all():
    """Save all documents without closing them."""

    (paths, _) = shared.get_paths()
    for path in paths:
        document = docxrev.Document(path, save_on_exit=True, close_on_exit=False)
        with document:
            pass



"""Update all grades."""

from __future__ import annotations

try:
    import docxrev
except TypeError as error:
    raise TypeError(
        "Cannot access the document. Save the document, close it,\n"
        "then re-open it and try again."
    ) from error

from gradedoc import shared
from gradedoc.update_grade import update_grade


def update_all_grades():
    """Update all grades."""

    (paths, gradebook_path) = shared.get_paths()

    for path in paths:
        document = docxrev.Document(path)
        update_grade(document, gradebook_path)



"""Toggle the review pane of the active document."""

from __future__ import annotations

try:
    import docxrev
except TypeError as error:
    raise TypeError(
        "Cannot access the document. Save the document, close it,\n"
        "then re-open it and try again."
    ) from error
from win32com.client import constants

from gradedoc import shared


def toggle_active_review_pane():
    """Toggle the review pane of the active document."""
    active_document = docxrev.get_active_document(save_on_exit=False)
    toggle_review_pane(active_document)


def toggle_review_pane(document: docxrev.Document):
    """Toggle the review pane of the document."""

    (paths, _) = shared.get_paths()

    with document:
        # Check if the document is in paths
        in_paths = document.path in paths  # we consume `paths` here
        # Now update the grade or raise an exception
        if in_paths:
            document.com.ActiveWindow.View.SplitSpecial = constants.wdPaneRevisions
        else:
            raise IOError("Active document not in paths.")



"""Update a document's grade as well as the gradebook."""
from __future__ import annotations

import csv
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Iterator, List

try:
    import docxrev
except TypeError as error:
    raise TypeError(
        "Cannot access the document. Save the document, close it,\n"
        "then re-open it and try again."
    ) from error
from natsort import natsorted

from gradedoc import shared
from gradedoc.configs import config
from gradedoc.toggle_active_review_pane import toggle_review_pane

__all__ = ["update_grade"]


def update_grade(document: docxrev.Document, gradebook_path: Path):
    """Update a document's grade as well as the gradebook.

    Parameters
    ----------
    document
        Document.
    gradebook_path
        Gradebook.
    """

    with document:
        grade = grade_document(document)
        update_document_scores(document, grade)
        update_gradebook(document, gradebook_path, grade)
        toggle_review_pane(document)


def grade_document(document: docxrev.Document) -> Grade:  # noqa: C901
    """Grade a document.

    Parameters
    ----------
    document
        Document.
    """

    # Search comments in reverse. This means that point deductions in any given section
    # will be encountered before we reach the header comment. This allows us to walk
    # through the comments once, and exit cleanly once we reach the first header
    # comment. We use `iter` and `next` on `comments` because they advance at a
    # different pace than the `header_comment_patterns` and `max_scores`.
    comments = iter(reversed(document.comments))
    header_comment_patterns = reversed(shared.HEADER_COMMENT_PATTERNS)
    max_scores = reversed(shared.MAX_SCORES)

    # Prepare the lists to be returned. They will be reversed at the end.
    scores: List[int] = []
    header_comments: List[docxrev.com.Comment] = []
    deductions = 0
    codes: List[str] = []

    # Get the scores for each section
    comment = safe_next(comments)
    for header_comment_pattern, max_score in zip(header_comment_patterns, max_scores):

        # Process point deductions while we don't recognize a header comment
        content_points_lost = 0
        while not header_comment_pattern.match(comment.text):

            # Increment the points lost in this section for matching comments
            if match := shared.CONTENT_POINTS_LOST_PATTERN.match(comment.text):
                content_points_lost += int(match["value"])

            # Increment total deductions for matching comments
            if match := shared.DEDUCTION_PATTERN.match(comment.text):
                deductions += int(match["value"])

            # Add text description to common deduction codes found in the document
            matches = [
                pattern.match(comment.text)
                for pattern in shared.COMMON_DEDUCTION_PATTERNS
            ]
            if any(matches):
                first_match = [match for match in matches if match][0]
                code = first_match["code"]
                if "D" not in code:  # Skip if this is a custom deduction
                    if not config.codes:
                        raise ValueError(
                            "No deduction codes found in configs."
                            "One config file may be overriding another,"
                            "make sure you don't have an empty 'codes' list."
                        )
                    try:
                        feedback = config.codes[code]
                    except KeyError as error:
                        raise KeyError(
                            f"No entry in configs for code {code}."
                        ) from error
                    first_line_of_comment = comment.text.split("\r", maxsplit=1)[0]
                    comment.update(first_line_of_comment + "\n\n" + feedback)
                    codes.append(code)

            # Try to get the next comment, raising an error if there are none left
            try:
                comment = safe_next(comments)
            except StopIteration as error:
                raise StopIteration(
                    "No comments in document, or template comments are out of order.\n"
                    f"Document: {document.name}"
                ) from error

        # We found a header comment. Store it and move on.
        header_comments.append(comment)
        comment = safe_next(comments)

        # Store the score for this section and skip the header comment
        scores.append(max_score - content_points_lost)

    header_comments.reverse()
    scores.reverse()
    return Grade(header_comments, scores, deductions, "; ".join(natsorted(codes)))  # type: ignore


def update_document_scores(document: docxrev.Document, grade: Grade):
    """Update document scores with the determined grade.

    Parameters
    ----------
    document
        Document.
    grade
        Grade.
    """

    # Update the summary comment scores
    summary_comment = document.comments[0]
    substitution = (
        rf"\g<content>{grade.content}\n"
        rf"\g<deductions>{grade.deductions}\n"
        rf"\g<grade>{grade.total}"
    )
    summary_comment.update(
        shared.SUMMARY_COMMENT_PATTERN.sub(substitution, summary_comment.text)
    )

    # Update the scores of the header comments
    for comment, pattern, score in zip(
        grade.header_comments, shared.HEADER_COMMENT_PATTERNS, grade.scores
    ):
        substitution = rf"\g<header>{score}"
        comment.update(pattern.sub(substitution, comment.text))


def update_gradebook(document: docxrev.Document, gradebook_path: Path, grade: Grade):
    """Write the grade for the paper being graded to a CSV file.

    Parameters
    ----------
    gradebook_path
        Gradebook.
    document
        Document.
    grade
        Grade.
    """

    # Prepare rows for the CSV
    header_row = [
        "Document",
        "Grade",
        "Total Content",
        "Total Deductions",
        "Deduction Codes",
        *[header.upper() for header in shared.FULL_HEADERS],
    ]

    new_row = [
        document.name,
        grade.total,
        grade.content,
        grade.deductions,
        grade.deduction_codes,
        *grade.scores,
    ]
    rows_to_write: List[Any] = []

    # Create the CSV and write the header if it doesn't exist
    if not os.path.exists(gradebook_path):
        with open(gradebook_path, "w", newline="") as file:
            writer = csv.writer(file)
            writer.writerow(header_row)

    # Build the rows of the CSV, overwriting or appending the current paper being graded
    with open(gradebook_path, "r", newline="") as file:

        update_existing_row = False
        reader = csv.reader(file)
        next(reader)  # skip the header

        # If the paper has already been graded, update that row, otherwise append it
        for row in reader:
            if row and new_row[0] == row[0]:
                rows_to_write.append(new_row)
                update_existing_row = True
            else:
                rows_to_write.append(row)
        if not update_existing_row:
            rows_to_write.append(new_row)

    # Write the updated file
    with open(gradebook_path, "w", newline="") as file:
        writer = csv.writer(file)
        writer.writerow(header_row)
        for row in rows_to_write:
            writer.writerow(row)


def safe_next(comments: Iterator[docxrev.com.Comment]) -> docxrev.com.Comment:
    """Safely get the next comment.

    Safely get the next comment, raising an error if all comments have been exhausted.

    Parameters
    ----------
    comments
        A comment iterator.
    document
        The document.

    Raises
    ------
    StopIteration
        If all comments are exhausted.
    """
    comment = None
    try:
        comment = next(comments)
    except StopIteration as error:
        if comment:
            message = (
                f"Exhausted comments in {comment.in_document.name} "
                "before finding all header comments."
            )
        else:
            message = "No comments in document."
        raise StopIteration(message) from error

    return comment


@dataclass
class Grade:
    """A grade."""

    header_comments: List[docxrev.com.Comment]
    """Header comments."""

    scores: List[int]
    """Scores."""

    deductions: int
    """Deductions."""

    @property
    def content(self) -> int:
        """Content score."""
        return sum(self.scores)

    @property
    def total(self) -> int:
        """Total score."""
        return self.content - self.deductions

    deduction_codes: str
    """ A semicolon delimited string of deduction codes."""



from contextlib import contextmanager

import pywintypes

try:
    import docxrev
except TypeError as error:
    raise TypeError(
        "Cannot access the document. Save the document, close it,\n"
        "then re-open it and try again."
    ) from error
import fire

from gradedoc.add_template_comments import add_template_comments
from gradedoc.close_all import close_all
from gradedoc.copy_example import copy_example
from gradedoc.delete_all_comments import delete_all_comments
from gradedoc.open_all import open_all
from gradedoc.save_all import save_all
from gradedoc.toggle_active_review_pane import toggle_active_review_pane
from gradedoc.update_active_grade import update_active_grade
from gradedoc.update_all_grades import update_all_grades


def main():
    with word():
        try:
            fire.Fire(
                {
                    "example": copy_example,
                    "active": update_active_grade,
                    "all": update_all_grades,
                    "addcom": add_template_comments,
                    "open": open_all,
                    "save": save_all,
                    "close": close_all,
                    "delcom": delete_all_comments,
                    "pane": toggle_active_review_pane,
                }
            )
        except pywintypes.com_error as error:
            raise TypeError(
                "Cannot access the document. Save the document, close it,\n"
                "then re-open it and try again."
            ) from error


@contextmanager
def word():
    try:
        yield
    finally:
        try:
            docxrev.quit_word_safely()  # If used as a CLI, quit Word if nothing was open
        except pywintypes.com_error as error:
            raise TypeError(
                "Cannot access the document. Save the document, close it,\n"
                "then re-open it and try again."
            ) from error



"""Update the grade of the active document."""

from __future__ import annotations

try:
    import docxrev
except TypeError as error:
    raise TypeError(
        "Cannot access the document. Save the document, close it,\n"
        "then re-open it and try again."
    ) from error

from gradedoc import shared
from gradedoc.update_grade import update_grade


def update_active_grade():
    """Update the active document grade, insert common comments, update gradebook."""

    (paths, gradebook_path) = shared.get_paths()

    active_document = docxrev.get_active_document(save_on_exit=False)
    with active_document:
        # Check if the document is in paths
        in_paths = active_document.path in paths  # we consume `paths` here
        # Now update the grade or raise an exception
        if in_paths:
            active_document.save_on_exit = True  # only save if we're updating the grade
            update_grade(active_document, gradebook_path)
        else:
            raise Exception("Active document not in paths.")



"""Delete all comments from all documents."""

from __future__ import annotations

try:
    import docxrev
except TypeError as error:
    raise TypeError(
        "Cannot access the document. Save the document, close it,\n"
        "then re-open it and try again."
    ) from error

from gradedoc import shared


def delete_all_comments():
    """Delete all comments from all documents."""

    response = input(  # nosec
        "Are you sure you want to delete all comments? [y/N] >>> "
    )
    if response.lower() == "y":
        (paths, _) = shared.get_paths()
        for path in paths:
            docxrev.Document(path).delete_comments()
    else:
        print("Not deleting comments.")



from pathlib import Path
from shutil import copytree

CWD = Path.cwd()


def copy_example(path: Path = CWD):
    """Copy example configuration and AutoHotkey scripts to the current directory."""
    import gradedoc

    scripts = Path(gradedoc.__file__).parent / "example"
    copytree(scripts, path / scripts.name)



"""Shared values."""


from __future__ import annotations

from pathlib import Path
import re
from typing import Tuple

from gradedoc.configs import config

# * -------------------------------------------------------------------------------- * #
# * PATHS

if docx_dir := config.get("docx_dir"):
    docx_dir = Path(docx_dir)
else:
    docx_dir = Path().cwd()

DOCX = r"[!~$]*.docx"  # excludes "~$" prefix temporary files
PATHS = list(docx_dir.resolve().glob(DOCX))
GRADEBOOK_NAME = "grades.csv"
GRADEBOOK_PATH = Path(docx_dir / GRADEBOOK_NAME)


# * -------------------------------------------------------------------------------- * #
# * COMMENTS
# Patterns here use named groups to be reused during replacement operations.

if headers := config.get("headers"):
    header_names = headers.keys()
    max_scores = headers.values()
else:
    header_names = [
        "abstract",
        "introduction and theory",
        "procedures",
        "results and discussion",
        "conclusion",
    ]
    max_scores = [10, 20, 20, 40, 10]

# For searching document content to insert comments
HEADERS = header_names

# For the actual content of the inserted comments
FULL_HEADERS = HEADERS
MAX_SCORES = max_scores

# The summary to be inserted and its pattern for later updating by regex
SUMMARY_COMMENT = "TOTAL CONTENT: 100\nTOTAL DEDUCTIONS: 0\nGRADE: 100/100"
# We match on "\r" because Microsoft Word implicitly converts "\n" to "\r" on insertion
SUMMARY_COMMENT_PATTERN = re.compile(
    r"(?P<content>TOTAL CONTENT: )-?\d+\r"
    r"(?P<deductions>TOTAL DEDUCTIONS: )-?\d+\r"
    r"(?P<grade>GRADE: )-?\d+"
)

# The header comments to be inserted
HEADER_COMMENTS = [
    f"{header.upper()}: {score}/{score}"
    for header, score in zip(FULL_HEADERS, MAX_SCORES)
]
# The pattern for finding header comments as well as substituting scores within them
HEADER_COMMENT_PATTERNS = [
    re.compile(rf"(?P<header>{header.upper()}: )-?\d+") for header in FULL_HEADERS
]


# * -------------------------------------------------------------------------------- * #
# * SCORING
# Patterns here use named groups to be reused during scoring operations.

# Matches comments with a number at the very start of the comment
CONTENT_POINTS_LOST = r"(?P<value>\d{1,3})"
CONTENT_POINTS_LOST_PATTERN = re.compile(CONTENT_POINTS_LOST)

# Matches comments with "D", then a number, at the very start of the comment
DEDUCTION = r"D(?P<value>\d{1,3})"
DEDUCTION_PATTERN = re.compile(DEDUCTION)

# Matches three types of common deduction patterns, e.g. "D2: G2", "2: A1", or "G12"
COMMON_DEDUCTION_CODE = r"(?P<code>[A-Z]\d{1,3})"
COMMON_DEDUCTION_PATTERNS = [
    re.compile(rf"{CONTENT_POINTS_LOST}:\s?{COMMON_DEDUCTION_CODE}"),
    re.compile(rf"{DEDUCTION}:\s?{COMMON_DEDUCTION_CODE}"),
    re.compile(COMMON_DEDUCTION_CODE),
]

# * -------------------------------------------------------------------------------- * #
# * FUNCTIONS


def get_paths() -> Tuple[list[Path], Path]:
    return PATHS, GRADEBOOK_PATH



"""Add template comments to all documents."""

from __future__ import annotations

try:
    import docxrev
except TypeError as error:
    raise TypeError(
        "Cannot access the document. Save the document, close it,\n"
        "then re-open it and try again."
    ) from error
from win32com.client import constants

from gradedoc import shared


def add_template_comments():
    """Add template comments to all documents."""

    (paths, _) = shared.get_paths()

    for path in paths:
        with docxrev.Document(path) as document:

            # Don't modify documents that already have comments
            if document.comments:
                continue

            # Convenient access to COM objects
            com_selection = document.com.Windows(1).Selection
            com_find = com_selection.Find

            document.com.Application.CommandBars.ExecuteMso("TableOfContentsRemove")

            # Go to the top of the document and add the summary comment
            com_selection.HomeKey(constants.wdStory)
            com_selection.Comments.Add(
                Range=com_selection.Range, Text=shared.SUMMARY_COMMENT
            )

            # Find headers and add header comments pointing to them
            for header, header_comment in zip(shared.HEADERS, shared.HEADER_COMMENTS):

                # Find the header
                com_find.Execute(
                    FindText=header,  # Find the header
                    # Ensure certain `Find` options are set
                    Forward=True,
                    MatchCase=False,
                    MatchWholeWord=False,
                    MatchWildcards=False,
                    MatchSoundsLike=False,
                    MatchAllWordForms=False,
                    Wrap=False,
                    Format=False,
                    Replace=constants.wdReplaceNone,
                )

                # Add the header comment pointing to the `Range` just found
                com_selection.Comments.Add(
                    Range=com_selection.Range,
                    Text=header_comment,
                )
                com_selection.EndKey()

            # Go back to the top of the document when done
            com_selection.HomeKey(constants.wdStory)



import csv

import yaml

with open("config.yaml") as file:
    all_codes = yaml.safe_load(file)

with open("codes.csv", "w", newline="") as file:
    writer = csv.writer(file, delimiter=";")
    for key, value in all_codes.items():
        writer.writerow([key, value])
